/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"

#include <ctype.h>
#include <pwd.h>
#include <sys/stat.h>

#ifdef HAVE_DIRENT_H
#include <dirent.h>
typedef struct dirent direntry;
#else
typedef struct direct direntry;
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#include "dict.h"
#include "sysdict.h"
#include "file.h"
#include "udb.h"
#include "str.h"
#include "stack.h"
#include "free.h"
#include "ps_part.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif


char *effective_user= NULL;	/* set by func_selectuser() */


/* FILE HANDLE LIST: Any file that any user is supposed to be able to edit
 * should be in the list below.  This tells where it is, who can edit it
 * and what type of file it is.  Only files on this list can be written,
 * but most any publically readable file can be read.
 */

#define TYP_TEXT   0	/* Regular Text File */
#define TYP_TAG    1	/* Tag File */

#define HD_HOME 0	/* in user's home directory */
#define HD_CFDR 1	/* in user's .cfdir or home directory */
#define HD_PART 2	/* in bbs/part directory */
#define HD_CONF 3	/* in current conference directory */
#define HD_SYST 4	/* in main bbs directory */

#if 0			/* Not using this */
#define HD_GROU 5	/* File lives in main bbs directory with group suffix */
#endif

#define AC_USER 0	/* Only the user can access this file */
#define AC_CONF 1	/* Only the fairwitness can access this file */
#define AC_GADM 2	/* Only group admin and sysadm can access this file */
#define AC_SYST 3	/* Only the sysadm can access this file */

/* These are indexes into the table below.  You must edit them when you edit
 * the table.  This stinks.
 */

#define FH_NONE			-1
#define FH_BACKTALK		 0
#define FH_CFLIST		 1
#define FH_CFONCE		 2
#define FH_CFRC			 3
#define FH_PLAN			 4
#define FH_BBSRC		 5
#define FH_BULL			 6
#define FH_CONFIG		 7
#define FH_CONFLIST		 8
#define FH_CONFMENU		 9
#define FH_CONFRC		10
#define FH_DFLT_BACKTALK	11
#define FH_DFLT_CFLIST		12
#define FH_GLIST		13
#define FH_INDEX		14
#define FH_LOGIN		15
#define FH_LOGOUT		16
#define FH_MOTD_HTML		17
#define FH_PUBLICTXT		18
#define FH_SECRET		19
#define FH_SETTINGS		20
#define FH_ULIST		21
#define FH_WELCOME		22

struct fnl {
	char *handle;	/* Name of file */
	int type;	/* Type of file */
	int loc;	/* Location of file */
	int access;	/* Who can write file */
	char *pu_flag;	/* Flag passed to partutil to permit file */
	int dflt;	/* File to read for default value */
	} fhand[]= {

	/* HANDLE          TYPE      LOCATION WRITER   PU_FLAG */
	/* -------------------------------------------------------- */
#ifdef PART_DIR
	{ ".backtalk",	   TYP_TAG,  HD_PART, AC_USER, NULL, FH_DFLT_BACKTALK },
	{ ".cflist",	   TYP_TEXT, HD_PART, AC_USER, NULL, FH_DFLT_CFLIST },
#else
	{ ".backtalk",	   TYP_TAG,  HD_CFDR, AC_USER, "-b", FH_DFLT_BACKTALK },
	{ ".cflist",	   TYP_TEXT, HD_CFDR, AC_USER, "-l", FH_DFLT_CFLIST },
#endif
	{ ".cfonce",	   TYP_TEXT, HD_CFDR, AC_USER, NULL, FH_NONE },
	{ ".cfrc",	   TYP_TEXT, HD_CFDR, AC_USER, NULL, FH_NONE },
	{ ".plan",	   TYP_TEXT, HD_HOME, AC_USER, "-n", FH_NONE },
	{ "bbsrc",	   TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "bull",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "config",	   TYP_TEXT, HD_CONF, AC_SYST, NULL, FH_NONE },
	{ "conflist",      TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "confmenu",      TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "confrc",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "dflt.backtalk", TYP_TAG,  HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "dflt.cflist",   TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "glist",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "index",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "login",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "logout",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "motd.html",	   TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "public.txt",	   TYP_TEXT, HD_SYST, AC_SYST, NULL, FH_NONE },
	{ "secret",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "settings",	   TYP_TAG,  HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "ulist",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },
	{ "welcome",	   TYP_TEXT, HD_CONF, AC_CONF, NULL, FH_NONE },

	{ NULL,		   0,	       0,       0,       NULL }};


/* CHECK_MODE - Check if file whose stats are in st is accessible to the
 * give uid and gid.  If uid or gid are -1, that indicates "no user".
 * Mode is some bitwise-or of 4 for read, 2 for write, 1 for execute.
 * We require all the types of access flagged.
 */

int check_mode(struct stat *st, long uid, long gid, int mode)
{
    if (st->st_uid == uid && uid != -1)
	return ((st->st_mode >> 6) & mode) == mode;
    else if (st->st_gid == gid && gid != -1)
	return ((st->st_mode >> 3) & mode) == mode;
    else
	return (st->st_mode & mode) == mode;
}


/* USER_ACCESS - Check if the user has read (mode=4), write (mode=2), or
 * execute (mode=1) access to the named file.  If mode is a bitwise-OR of
 * these, all the flagged access is required.  If dir is true, it must be
 * a directory.  Like access(), this returns zero if he has the required
 * access.  Files that don't exist are not readable or executable, but
 * they may be writable if the parent directory is writable.  If "sys" is
 * true, then we only check if backtalk has the desired access, not if the
 * current user has it.  Return codes are:
 *     0 = we have the required access.
 *     1 = we don't have the required access, but the file exists.
 *     2 = we don't have the required access, and the file does not exist.
 */

#ifdef REAL_UID
#define UA_UID    idict(VAR_UID)
#else
#define UA_UID    -1
#endif
#ifdef REAL_GID
#define UA_GID   idict(VAR_GID)
#else
#define UA_GID   -1
#endif

int user_access(char *fname, int mode, int dir, int sys)
{
    struct stat st;
    char *c;
    int rc;

    if (stat(fname,&st))
    {
	if ((mode != 2 && mode != 6) || dir) return 2;
	/* Writing a file that doesn't exist -- check parent directory */
	if ((c= strrchr(fname,'/')) == NULL)
	    rc= stat(".",&st);
	else
	{
	    *c= '\0';
	    rc= stat(fname,&st);
	    *c= '/';
	}
	if (rc) return 2;
	/* If we are going to create a file, need write/execute access */
	mode= 3;
    }
    else if (mode == 0)
	return 0;

    if (dir && !(st.st_mode & S_IFDIR)) return 1;

    return (((sys || check_mode(&st, UA_UID, UA_GID, mode)) &&
	    check_mode(&st, geteuid(), getegid(), mode)) ? 0 : 1 );
}


/* EFFECTIVE_DIR - Return the home directory of the given user.  If the
 * user is null, then return the currently logged in user.  Often the
 * user is the effective_user.
 */

char *effective_dir(char *user)
{
    if (user == NULL)
	return sdict(VAR_HOMEDIR);
    else
    {
	/* We don't call gethomedir here because we already called getdbname
	 * when we did "selectuser" so this data is in cache and this is
	 * actually faster than gethomedir().
	 */
	struct passwd *pw= getdbnam(user);
	return pw->pw_dir;
    }
}


/* HAND_PATH - Write the full pathname of the i-th handle into the buffer
 * pointed to by path.  User is NULL for the current user.  Return true on
 * success.
 */

int hand_path(int i, char *user, char *path)
{
    char *fn= (i == FH_BBSRC || i == FH_CONFRC) ? "rc" : fhand[i].handle;
    char *e;

    switch(fhand[i].loc)
    {
    case HD_CONF:
	sprintf(path,"%s/%s", sdict(VAR_CONFDIR), fn);
	break;

#ifdef HD_GROU
    case HD_GROU:
        e= user_gname();
	if (e == NULL) return 0;
	sprintf(path,"%s/%s.%s", sdict(VAR_BBSDIR), fn, e);
	break;
#endif

    case HD_SYST:
	sprintf(path,"%s/%s", sdict(VAR_BBSDIR), fn);
	break;

    case HD_HOME:
	sprintf(path,"%s/%s", effective_dir(user), fn);
	break;

    case HD_CFDR:
	homepartpath(path, effective_dir(user), fn);
	break;

#ifdef PART_DIR
    case HD_PART:
	bbspartpath(path, user ? user : sdict(VAR_ID), fn);
	break;
#endif

    default:
	die("Unknown location code %d",fhand[i].loc);
    }
    return 1;
}


/* CHECK_HANDLE - Expand a handle or filename into a regular filename.  If it
 * starts with a '*' it is a handle and is looked up in the table. If it isn't
 * there, or the type doesn't match the expected type, we die with an error
 * message.  If we are reading a text file, and it isn't accessible, it may
 * return the pathname of a readable default file instead.  With tag files,
 * you only get the dflt file if you request it.  Otherwise store the full path
 * name in the given buffer.  If the handle doesn't start with a '*', we just
 * copy it, with no checking, so long as we don't intend to do a write.
 * "Mode" is a boolean-or of:
 *		1 = check execute access
 *		2 = check write access
 *		4 = check read access
 *		8 = don't check file type
 *             16 = force open of dflt file, instead of file itself.
 *	       32 = check if it is a directory
 * It bombs on illegal requests, and returns 0 for success, and 1 for legal
 * requests for files that aren't accessible (eg, don't exist).
 */

int check_handle(char *handle, int exptype, int mode, char *path)
{
    static int last= -1;
    int i, rc;
    char *e;

    if (handle == NULL || handle[0] == '*')
    {
	if (mode&32) return 1;	/* No handles are for directories */
	if (handle == NULL)
	{
	    /* Reuse previous handle */
	    if (last == -1) die("no previous check_handle");
	    i= last;
	}
	else
	{
	    /* Search for the handle */
	    for (i= 0; fhand[i].handle != NULL; i++)
		if (!strcmp(handle+1,fhand[i].handle))
		{
		    last= i;
		    goto found;
		}

	    if (mode&8) return 1;
	    die("Illegal file handle: %s",handle);
found:;
	}

	/* Do default file instead */
	if (mode&16)
	{
	    if (fhand[i].dflt == FH_NONE)
		return 1;
	    i= fhand[i].dflt;
	}

	/* Check type correctness */
	if (!(mode&8))
	{
	    if (exptype == TYP_TEXT && fhand[i].type == TYP_TAG)
		die("%s is not a text file",handle);
	    if (exptype == TYP_TAG && fhand[i].type != TYP_TAG)
		die("%s is not a tag file",handle);
	}

	/* Check if we have access */
	switch(fhand[i].access)
	{
	case AC_CONF:
	    if ((mode&2) && !bdict(VAR_AMFW))
	    {
		if (mode&8) return 1;
		die("Only fairwitnesses may write %s",handle);
	    }
	    break;

	case AC_SYST:
	    if ((mode&2) && !bdict(VAR_AMADM))
	    {
		if (mode&8) return 1;
		die("Only conference adminstrator may write %s",handle);
	    }
	    break;

	case AC_USER:
	    break;

	default:
	    die("Unknown access code %d",fhand[i].loc);
	}

	/* Construct full path name */
	if (hand_path(i,effective_user,path))
	{
	    /* If it exists and we have access, we are done */
	    if (!(rc= user_access(path, mode&7, 0, 1))) return 0;

	    /* File doesn't exist, or we don't have access */
#ifdef PARTUTIL
	    /* See if partutil can be used to permit this file to us
	     * This is only worth doing if the file exists or we are
	     * planning to write to it.  Rather than read a non-existant
	     * file, we'd prefer to use the default.
	     */
	    if ((rc == 1 || (mode&2)) && fhand[i].access == AC_USER &&
		fhand[i].pu_flag != NULL)
	    {
		partutil(fhand[i].pu_flag,NULL,sdict(VAR_ID));
		if (!user_access(path, mode&7, 0, 1)) return 0;
	    }
#endif /*PARTUTIL*/
	}

	/* See if there is a default file to read instead */
	while (fhand[i].dflt != FH_NONE && exptype == TYP_TEXT && !(mode&2))
	{
	    i= fhand[i].dflt;
	    if (hand_path(i,effective_user,path) &&
	        !user_access(path, mode&7, 0, 1)) return 0;
	}
	return 1;
    }
    else
    {
	/* No default file */
	if (mode&16) return 1;

	/* Normal filenames */
	if (mode&2)
	{
	    if (mode&8) return 1;
	    die("Filenames must be given as handles when writing");
	}
	strcpy(path,handle);

	if (user_access(path, mode&7, mode&32, 0))
	    return 1;
    }
    return 0;
}


/* FUNC_READ:  <filehandle> read <string>
 * Push the entire contents of the named file as a single string onto the
 * stack.  The name can be a normal file path, or a handle starting with
 * a star.  If the file cannot be read, returns a {} string.
 */

void func_read()
{
    char *data, *hand;
    char fname[BFSZ];
    int err;

    err= check_handle(hand= peek_string(), TYP_TEXT, 4, fname);
    free(hand);

    if (err || (data= read_file(fname)) == NULL)
	data= strdup("");

    repl_top((void *)data);
}


/* FUNC_WRITE:  <string> <filehandle> write -
 * Overwrite or create the named file with the given string as it's contents.
 * The name must be a handle.
 */

void func_write()
{
    struct stat st;
    char *fhandle=pop_string();
    char fname[BFSZ];
    char *data, *d;
    FILE *fp;
    int err;
    int is_secret;

    err= check_handle(fhandle, TYP_TEXT, 2, fname);
    is_secret= !strcmp(fhandle,"*secret");
    free(fhandle);

    if (err || (fp= fopen(fname,"w")) == NULL || fstat(fileno(fp),&st))
	    die("Cannot open file %s to write",fname);
    if (is_secret) fchmod(fileno(fp),0600);

    data= pop_string();

    for (d= data ; d[0] != '\0'; d++ )
	if (d[0] != '\r' || d[1] != '\n')
	    fputc(*d,fp);
    if (d > data && d[-1] != '\n')
	fputc('\n',fp);

    fclose(fp);
    free(data);
}


/* FUNC_LINE:	<filehandle> <number> line <string>
 * Read the given line number from the given file.  The first line is number
 * zero.  Returns null strings if the file does not exist, is not readable,
 * or is not that long.  May not work right with files with extremely long
 * lines.  The name may be either a regular filename or a handle starting with
 * a star.
 */

void func_line()
{
    int n= pop_integer();
    char *fhandle= peek_string();
    char fname[BFSZ];
    char bf[BFSZ+1];

    if (check_handle(fhandle, TYP_TEXT, 4, fname))
	die("Cannot read %s",fhandle);
    free(fhandle);

    nth_line(fname,n,bf,FALSE);
    push_string(bf,TRUE);
}


/* FUNC_PATHNAME:  <filehandle> <user> pathname <path>
 * Construct the full pathname for a file handle would expand to for a given
 * user.  If user is an empty string, the current user is used.
 */

void func_pathname()
{
    char *user= pop_string();
    char *fhandle= pop_string();
    char bf[BFSZ+1];
    int i;

    if (user[0] == '\0') user= NULL;

    /* Search for the handle */
    for (i= 0; fhand[i].handle != NULL; i++)
	if (!strcmp(fhandle+1,fhand[i].handle))
	    break;
    if (fhand[i].handle == NULL)
	die("Unknown file handle %s",fhandle);

    hand_path(i,user,bf);

    push_string(bf,TRUE);
}


/* F_NTH_LINE - this is like nth_line() but works on an already open file */

int f_nth_line(FILE *fp, int n, char *bf, int skip)
{
    int i;

    for (i= 0; i <= n; i++)
    {
	if (fgets(bf,BFSZ,fp) == NULL)
	{
	    bf[0]= '\0';
	    return 0;
	}
	if (skip && (bf[0] == '#' || bf[0] == '\n'))
		i--;
    }
    *firstin(bf,"\n")= '\0';
    bf[BFSZ]= '\0';
    return 0;
}
    

/* NTH_LINE - read the nth line from the named file, returning it in the
 * named size BFSZ+1 buffer.  If skip is true, lines with #'s in front and
 * blank lines are not counted.  First line is number 0.  If the line or the
 * file does not exist, a null string is returned.  Returns 1 if the file
 * does not exist, 0 if it does.  If the line number is too high, a null
 * string is returned.
 */

int nth_line(char *fname, int n, char *bf, int skip)
{
    int i,rc;
    FILE *fp;

    if ((fp= fopen(fname,"r")) == NULL)
    {
	bf[0]= '\0';
	return 1;
    }
    rc= f_nth_line(fp, n, bf, skip);
    
    fclose(fp);
    return rc;
}


/* NEXTDIR - Return the next filename in curr_direct.  Data is in static
 * storage that will be overwritten by the next call.  It closes curr_direct
 * when you hit the end.
 */

DIR *curr_direct= NULL;

char *nextdir()
{
    direntry *dir;
    static char none[]= "";

    if ((dir= readdir(curr_direct)) == NULL)
    {
	closedir(curr_direct);
	curr_direct= NULL;
    	return none;
    }
    else
    	return dir->d_name;
}


/* FUNC_FIRSTDIR:  <directory> firstdir <filename>
 * Return the first filename in a directory.  If the directory name is an
 * empty string, just close any previously open directory.  Bomb if the
 * directory doesn't exist or isn't readable.  Returns an empty string
 * in the nearly impossible case of the directory being empty.
 */

void func_firstdir()
{
    char *dhandle= pop_string();
    char dname[BFSZ];

    if (curr_direct != NULL)
    {
    	closedir(curr_direct);
	curr_direct= NULL;
    }

    if (*dhandle == '\0')
    	return;

    if (check_handle(dhandle,0,32|8|4|1,dname))
    	die("handle '%s' does not refer to an accessable directory", dhandle);
    free(dhandle);

    if ((curr_direct= opendir(dname)) == NULL)
    	die("cannot open directory %s", dname);

    push_string(nextdir(), TRUE);
}


/* FUNC_NEXTDIR:  - nextdir <filename>
 * Return the next filename the directory selected by the last firstdir call.
 * Return an empty string if we hit the end of the directory.  Bomb if there
 * wasn't a previous call to firstdir, or if a previous call to nextdir
 * already hit the end.
 */

void func_nextdir()
{
    if (curr_direct == NULL)
    	die("no directory scan in progress");
    push_string(nextdir(), TRUE);
}


/* FUNC_DIRECTORY:  <filehandle> directory <bool>
 * Check if a file is a directory with read/execute access.
 */

void func_directory()
{
    char *dhandle= pop_string();
    char dname[BFSZ];

    push_integer(!check_handle(dhandle,0,32|8|4|1,dname));
    free(dhandle);
}


/* FUNC_READABLE:  <filehandle> readable <bool>
 * Check if a file is readable.
 */

void func_readable()
{
    char *fhandle= pop_string();
    char fname[BFSZ];

    push_integer(!check_handle(fhandle,0,8|4,fname));
    free(fhandle);
}


/* FUNC_WRITABLE:  <filehandle> writable <bool>
 * Check if a file is writable.  Only handles are ever writable.
 */

void func_writable()
{
    char *fhandle= pop_string();
    char fname[BFSZ];

    push_integer(!check_handle(fhandle,0,8|2,fname));
    free(fhandle);
}


/* FUNC_EXISTS:  <filehandle> exists <bool>
 * Check if a file exists.
 */

void func_exists()
{
    char *fhandle= pop_string();
    char fname[BFSZ];

    push_integer(!check_handle(fhandle,0,8|0,fname));
    free(fhandle);
}


/* FUNC_FILEDATE: <filehandle> filedate <time>
 * Returns time that the named file was last modified, or 0 if it does not
 * exist.
 */

void func_filedate()
{
    struct stat st;
    char *fhandle= pop_string();
    char fname[BFSZ];

    if (check_handle(fhandle, 0, 8|0, fname) || stat(fname,&st))
	push_time(0);
    else
	push_time(st.st_mtime);
    free(fhandle);
}


#if 0
/* FUNC_CHDIR: <dirname> chdir -
 * Change the current directory to the one named.
 */

void func_chdir()
{
    char *dir= pop_string();

    if (chdir(dir))
	die("could not change directory to %s",dir);
    free(dir);
}
#endif


/* POP_LITARRAY - Get an array of literals off the stack.  If the value on the
 * stack is just a single literal, form it into an array of one literal and
 * return that array.
 *
 * What is this doing in the file.c source module?
 */

void pop_litarray(Token *t)
{
    int i;

    pop_any(t);
    if (type(*t) == TK_UNBOUND_LITERAL || type(*t) == TK_BOUND_LITERAL)
    {
	/* Turn literal into an array of size one containing it */
	Array *a= (Array *)malloc(sizeof(Array));
	a->a= (Token *)malloc(sizeof(Token));
	a->a[0]= *t;
	a->sz= a->lk= 1;
	t->flag= TK_ARRAY;
	t->val= (void *)a;
    }
    else if (type(*t) == TK_ARRAY)
    {
	Token *ta= aval(*t)->a;

	/* Check each element to make sure it is a string */
	for (i= 0; i < aval(*t)->sz; i++)
	    if (type(ta[i]) != TK_UNBOUND_LITERAL &&
	        type(ta[i]) != TK_BOUND_LITERAL)
		die ("expected array containing only literals");
    }
    else
	die("Expected literal or array");
}


/* VARCMP - check if the given pattern matches the given variable name.
 * Any pattern ending with a dot matches all variables that start with the
 * same characters.  Patterns not ending with a dot need to exactly match
 * the variable name.
 *
 * Return values are *not* like strcmp().  We return 1 if they match, zero
 * otherwise.
 */

int varcmp(char *pat, char *var)
{
    char *p;

    for (p= pat; *p == *var; p++, var++)
	if (*p == '\0') return 1;

    return (p != pat && *p == '\0' && p[-1] == '.');
}


/* DO_LOADVAR - given an array and a filename, do the loadvar of those
 * variables from that file.  If varray is NULL, load all variables.
 */

int do_loadvar(Array *varray, char *fname)
{
    Token st;
    char *bf, *val, *end, *var;
    int i,is_string;

    if ((bf= read_file(fname)) == NULL)
    {
	/* Failure - do nothing */
	return 1;
    }

    /* Scan each line of the file */
    var= bf;
    while (*var != '\0')
    {
	/* Parse file line */
	if (*var == '*') var++;
	if ((val= strchr(var,'=')) == NULL)
		die("bad tag syntax");

	is_string= (val[1] == '(');

	if (is_string) {
	    int esc= 0;
	    for (end=val+1;*end != '\0' && (esc || *end != ')');end++)
	    	esc= (*end == '\\' && !esc);
	    if (end[0] == '\0' || end[1] != '\n' )
		die("bad string tag syntax");
	    end++;
	}
	else
	{
	    end= strchr(var,'\n');
	    if (end == NULL || val > end)
		die("bad integer tag syntax");
	}
	*(val++)= '\0';
	*end= '\0';


	/* Scan through array, to see if we should load this variable */
	if (varray != NULL)
	{
	    for (i= 0; i < varray->sz; i++)
	    {
		if (type(varray->a[i]) == TK_UNBOUND_LITERAL)
		{
		    if (varcmp(sval(varray->a[i]), var))
			break;
		}
		else
		{
		    if (varcmp(sysdict[ival(varray->a[i])].key, var))
		    	break;
		}
	    }
	}

	/* Load the variable */
	if (varray == NULL || i < varray->sz)
	{
	    /* Set variable to value from file */
	    if (is_string)
	    {
		val++;
		*(end-1)= '\0';
		unesctag(val, val, 0);
		st.flag= TK_STRING | TKF_FREE;
		st.val= (void *)strdup(val);
	    }
	    else if (isascii(*val) && isdigit(*val))
	    {
		st.flag= TK_INTEGER;
		st.val= (void *)atoi(val);
	    }
	    else
		die("%s tag has bad value",var);

	    if (type(varray->a[i]) == TK_UNBOUND_LITERAL)
		store_dict(var,&st,FALSE);
	    else
		set_dict(sysdict+ival(varray->a[i]),&st,0);
	}

	/* Go on to next line */
	var= end+1;
    }

    free(bf);
    return 0;
}


/* FUNC_LOADVAR:  <filehandle> <vars> loadvar <error>
 * Read varible settings for the listed variables from a file.  <vars> can be
 * either a single variable name or an array of variable names.  Pushs a
 * boolean which is true if the file was not readable.
 */

void func_loadvar()
{
    Token vt;
    char fname[BFSZ], *hand;
    int err,rc1= 1, rc2= 1;

    /* Get the variable name array */
    pop_litarray(&vt);

    /* Do the default version of the file first, if there is one */
    err= check_handle(hand= pop_string(), TYP_TAG, 16|4, fname);
    if (!err) rc1= do_loadvar(aval(vt),fname);

    /* Do the file itself, if it exists */
    if (!check_handle((hand[0]=='*') ? NULL : hand, TYP_TAG, 4, fname))
	rc2= do_loadvar(aval(vt),fname);

    free_val(&vt);
    free(hand);

    push_integer(rc1&rc2);
}

/* WRTSTRTAG - Write a string as a tag line */

void wrtstrtag(FILE *fp, char *name, char *value)
{
    fprintf(fp,"%s=(", name);
    while (*value != '\0')
    {
	switch (*value)
	{
	case '\r':                   break;
	case '\n': fputs("\\n",fp);  break;
	case '\\': fputs("\\\\",fp); break;
	case '(':  fputs("\\(",fp);  break;
	case ')':  fputs("\\)",fp);  break;
	default:   fputc(*value,fp); break;
	}
	value++;
    }
    fputs(")\n",fp);
}


/* WRTTAG - Write the token as a tag line */

void wrttag(FILE *fp, char *name, Token *t)
{
    if (type(*t) == TK_STRING)
	wrtstrtag(fp,name,sval(*t));
    else if (type(*t) == TK_INTEGER)
	fprintf(fp,"%s=%d\n", name, ival(*t));
    else
	die("%s is not a string or integer variable", name);
}


/* DO_SAVEVAR - this does most of the work for func_savevar below.  fname is
 * the filename to save in (all security checks are assumed to have been made
 * already), and vt is a pointer to an array of literal names.
 */

void do_savevar(char *fname, Token *vt)
{
    Token  *newt, *va;
    char *bf, *val, *end, *var;
    FILE *fp;
    int i, protect;

    va= aval(*vt)->a;

    /* Load file into memory */
    bf= read_file(fname);

    /* Truncate and prepare to rewrite */
    if ((fp= fopen(fname,"w+")) == NULL)
	die("Cannot open file %s to write",fname);

    /* Process each line of saved file contents, writing back or replacing */
    var= bf;
    while (var != NULL && *var != '\0')
    {
	/* Parse file line */
	if (protect= (*var == '*')) var++;
	val= strchr(var,'=');
	end= strchr(var,'\n');
	if (val != NULL && end != NULL && val > var && end > val)
	{
	    *(val++)= '\0';
	    *end= '\0';

	    /* Scan through variable names */
	    newt= NULL;
	    for (i= 0; i < aval(*vt)->sz; i++)
	    {
		if (type(va[i]) == TK_UNBOUND_LITERAL)
		{
		    if (!strcmp(sval(va[i]), var))
		    {
			if ((newt= get_dict(var)) == NULL)
			    die("variable %s not defined",var);
			sval(va[i])[0]= '\0';	/* Mark name saved */
			break;
		    }
		}
		else
		{
		    if (ival(va[i]) >= 0 &&
		        !strcmp(sysdict[ival(va[i])].key, var))
		    {
			newt= &(sysdict[ival(va[i])].t);
			va[i].val= (void *)-1;	/* Mark name saved */
			break;
		    }
		}
	    }
	    if (protect || newt == NULL)
		fprintf(fp,"%s=%s\n", var-protect, val);
	    else
		wrttag(fp,var,newt);
	}
	var= end+1;
    }

    /* Process variables not previously in file */
    for (i= 0; i < aval(*vt)->sz; i++)
    {
	if (type(va[i]) == TK_UNBOUND_LITERAL)
	{
	    if (sval(va[i])[0] != '\0')
	    {
		if ((newt= get_dict(sval(va[i]))) == NULL)
		    die("variable %s not defined",sval(va[i]));
		wrttag(fp, sval(va[i]), newt);
	    }
	}
	else
	{
	    if (ival(va[i]) >= 0)
		wrttag(fp, sysdict[ival(va[i])].key, &sysdict[ival(va[i])].t);
	}
    }

    fclose(fp);

    if (bf != NULL) free(bf);
}


/* FUNC_SAVEVAR:  <filehandle> <vars> savevar -
 * Save the current values of the listed variables into a file.  <vars> can be
 * either a single variable name or an array of variable names.  If the file
 * already contains data, old values for these variables will be replaced,
 * and new ones will be appended.
 */

void func_savevar()
{
    Token vt;
    char fname[BFSZ],*hand;

    /* Get the variable name array */
    pop_litarray(&vt);

    /* Get the file name */
    if (check_handle(hand= pop_string(), TYP_TAG, 2, fname))
	die("Cannot read %s",hand);
    free(hand);

    do_savevar(fname, &vt);

    free_val(&vt);
}


/* OPEN_SCRIPT - Open a script given the filename of either the compiled or
 * uncompiled script.  In the future, this should check if the script needs
 * recompilation, and do it if necessary.  It returns and file descriptor on
 * success, or -1 on failure.  If bname is non-null, it points to a BFSZ
 * storage area in which to return a copy of the binary file name.
 */

int open_script(char *fname, char *bname)
{
    char *sfile, *bfile, *x;
    int fd;
    int len= strlen(fname);

    if (len < 4 || strcmp(fname+len-3,".bt") != 0)
    {
	/* Name given had no .bt extension */
	sfile= x= (char *)malloc(len+4);
	strcpy(sfile,fname);
	strcpy(sfile+len,".bt");
	bfile= fname;
    }
    else
    {
	/* Name given had a .bt extension */
	bfile= x= (char *)malloc(len-2);
	strncpy(bfile,fname,len-3);
	bfile[len-3]= '\0';
	sfile= fname;
    }

    /* Future checks go here */

    /* Open the file */
    fd= open(bfile,0);

    /* Return a copy of the binary name */
    if (bname != NULL)
    {
	strncpy(bname,bfile,BFSZ-1);
	bname[BFSZ-1]= '\0';
    }

    /* Deallocate whichever name we had in allocated memory */
    free(x);

    return fd;
}


/* GET_SETTINGS - This loads certain settings from the conference
 * settings file.  It returns:
 *
 *      new_isel   - item selector for items which should be new on
 *                   first joining.  Defaults to "0,$" if not defined.
 *                   Should point to a BFSZ buffer.
 *
 *      new_since  - if positive, items that have had new responses in
 *                   this number of days should be new on first joining.
 *                   Defaults to 0 if not defined.
 *
 * Any other tags in the conference settings file are ignored.
 */

void get_settings(char *confdir, char *new_isel, int *new_since)
{
    char bf[BFSZ+1];
    char *var, *val, *end;
    FILE *fp;
    char ch;

    /* Set defaults */
    strcpy(new_isel,"1,$");
    *new_since= 0;

    sprintf(bf,"%s/settings",confdir);
    if ((fp= fopen(bf,"r")) == NULL) return;

    while (fgets(bf, BFSZ, fp) != NULL)
    {
    	var= bf + (bf[0] == '*');
    	if ((val= strchr(var, '=')) == NULL)
    	{
	    fclose(fp);
	    return;
	}
	*(val++)= '\0';
	end= val + strlen(val) - 1;
	if (*end == '\n')
	    *(end--)= '\0';
	else
	    /* skip over rest of over-long line */
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
	    	;

	if (!strcmp(var, "new_isel"))
	{
	    if ((*val == '(' || *val == '{') &&
	        (*end == ')' || *end == '}'))
	    {
		*end= '\0';
		unesctag(val+1,new_isel,0);
	    }
	}
	else if (!strcmp(var, "new_since"))
	    *new_since= atoi(val);
    }

    fclose(fp);
}


/* OPEN_CFLIST - open the the .cflist file.  If it doesn't exist, get the
 * dflt.cflist file instead.  Returns the open file descriptor if we got it,
 * or NULL if there was no accessible one.
 */

FILE *open_cflist(int mine)
{
    FILE *fp;
    char fname[BFSZ+1];

    if (check_handle("*.cflist",TYP_TEXT,mine ? 4 : 20,fname))
	return NULL;

    return fopen(fname,"r");
}
